@@ -226,6 +226,43 @@ class CommandTrnasHandler():
             return True
         else:
             return False
+        
+    def cmd_v3_flash_earse_all(self):
+        self._put_packet(alp.Command.FLASH_EARSE_ALL, b'')
+        time.sleep(3)
+        rep = self._get_packet()
+        if rep['command'] == alp.Command.FLASH_EARSE_ALL and rep['data'][0] == 0:
+            return True
+        else:
+            return False
+
+    def cmd_v2_ext_flash_write(self, page_addr, data):
+        paylod = page_addr.to_bytes(4, 'little') + data
+        self._put_packet(alp.Command.EXT_FLASH_WRITE, paylod)
+        rep = self._get_packet()
+        if rep['command'] == alp.Command.EXT_FLASH_WRITE and rep['data'][0] == 0:
+            return True
+        else:
+            return False
+
+    def cmd_v2_ext_flash_read(self):
+        pass
+
+    def cmd_v2_ext_flash_earse_all(self):
+        self._put_packet(alp.Command.EXT_FLASH_EARSE_ALL, b'')
+        rep = self._get_packet()
+        if rep['command'] == alp.Command.EXT_FLASH_EARSE_ALL and rep['data'][0] == 0:
+            return True
+        else:
+            return False
+
+    def cmd_v2_prog_ext_to_int(self):
+        self._put_packet(alp.Command.PROG_EXT_TO_INT, b'')
+        rep = self._get_packet()
+        if rep['command'] == alp.Command.PROG_EXT_TO_INT and rep['data'][0] == 0:
+            return True
+        else:
+            return False
 
     def cmd_v2_eep_set_pgsz(self, size):
         self._put_packet(alp.Command.EEPROM_SET_PGSZ,
@@ -294,20 +331,25 @@ class Loader():
     _device_name = str()
 
     _is_flash_prog = bool()
+    _is_ext_flash_prog = bool()
+    _is_ext_to_int = bool()
     _is_eeprom_prog = bool()
     _is_go_app = bool()
 
     _flash_file = str()
+    _ext_flash_file = str()
     _eep_file = str()
     _go_app_delay = int()
 
     class _Stage(enum.IntEnum):
         """燒錄狀態物件
         """
-        PREPARE = 0   # 準備中，偵測檔案、檢查參數等
-        FLASH_PROG = 1  # flash 燒錄中
-        EEP_PROG = 2    # eeprom 燒錄中
-        END = 3         # 結束，發送結束命令或跳轉應用命令
+        PREPARE = 0         # 準備中，偵測檔案、檢查參數等
+        FLASH_PROG = 1      # flash 燒錄中
+        EEP_PROG = 2        # eeprom 燒錄中
+        EXT_FLASH_PROG = 3  # ext-flash 燒錄中
+        EXT_TO_INT = 4      # ext to internel 燒錄中
+        END = 5             # 結束，發送結束命令或跳轉應用命令
 
     _stage = _Stage(_Stage.PREPARE)
     _stage_iter = None
@@ -316,12 +358,15 @@ class Loader():
     _cur_step = 0
 
     _flash_pages = list()
-    _eep_pages = list()
     _flash_page_idx = int()
+    _ext_flash_pages = list()
+    _ext_flash_page_idx = int()
+    _eep_pages = list()
     _eep_page_idx = int()
 
     # output info
     _flash_size = int(0)
+    _ext_flash_size = int(0)
     _eep_size = int(0)
     _prog_time = float(0)
 
@@ -330,19 +375,24 @@ class Loader():
         ser: serial.Serial,
         device_type: int = 0,
         is_flash_prog: bool = False,
+        is_ext_flash_prog: bool = False,
         is_eeprom_prog: bool = False,
+        is_ext_to_int: bool = False,
         is_go_app: bool = False,
         flash_file: str = '',
+        ext_flash_file: str = '',
         eeprom_file: str = '',
         go_app_delay: int = 0,
     ):
-        """初使化函式
+        """初始化函式
 
         Args:
             ser (serial.Serial): 通訊用的serial物件，須由外部先行開啟。
             device_type (int, optional): 燒錄裝置種類。預設為0。
             is_flash_prog (bool, optional): 是否要燒錄flash。預設為False。
+            is_ext_flash_prog (bool, optional): 是否要燒錄外部flash。預設為False。
             is_eeprom_prog (bool, optional): 是否要燒錄eeprom。預設為False。
+            is_ext_to_int (bool, optional): 是否要從外部flash燒錄檔案至內部flash。預設為False。
             is_go_app (bool, optional): 燒錄完是否執行應用。預設為False。
             flash_file (str, optional): 要燒錄的flash檔案。預設為''。
             eeprom_file (str, optional): 要燒錄的eeprom檔案。預設為''。
@@ -354,9 +404,12 @@ class Loader():
 
         self._device_type = device_type
         self._is_flash_prog = is_flash_prog
+        self._is_ext_flash_prog = is_ext_flash_prog
         self._is_eeprom_prog = is_eeprom_prog
+        self._is_ext_to_int = is_ext_to_int
         self._is_go_app = is_go_app
         self._flash_file = flash_file
+        self._ext_flash_file = ext_flash_file
         self._eep_file = eeprom_file
         self._go_app_delay = go_app_delay
 
@@ -384,6 +437,10 @@ class Loader():
     def flash_size(self):
         return self._flash_size
 
+    @property
+    def ext_flash_size(self):
+        return self._ext_flash_size
+
     @property
     def eeprom_size(self):
         return self._eep_size
@@ -407,6 +464,10 @@ class Loader():
             if os.path.isfile(self._flash_file) is False:
                 raise FileNotFoundError
 
+        if self._is_ext_flash_prog:
+            if os.path.isfile(self._ext_flash_file) is False:
+                raise FileNotFoundError
+
         if self._is_eeprom_prog:
             if os.path.isfile(self._eep_file) is False:
                 raise FileNotFoundError
@@ -415,7 +476,12 @@ class Loader():
             if self._go_app_delay > 65535:
                 raise exceptions.GoAppDelayValueError(self._go_app_delay)
 
+        if self._is_ext_to_int:
+            if self._go_app_delay > 65535:
+                raise exceptions.GoAppDelayValueError(self._go_app_delay)
+
         self._prepare_flash()
+        self._prepare_ext_flash()
         self._prepare_eeprom()
         self._prepare_device()
 
@@ -424,9 +490,14 @@ class Loader():
         if self._is_flash_prog:
             stg_list.append(self._Stage.FLASH_PROG)
             self._total_steps += len(self._flash_pages)
+        if self._is_ext_flash_prog:
+            stg_list.append(self._Stage.EXT_FLASH_PROG)
+            self._total_steps += len(self._ext_flash_pages)
         if self._is_eeprom_prog:
             stg_list.append(self._Stage.EEP_PROG)
             self._total_steps += len(self._eep_pages)
+        if self._is_ext_to_int:
+            stg_list.append(self._Stage.EXT_TO_INT)
         stg_list.append(self._Stage.END)
         self._total_steps += 1
         self._stage_iter = iter(stg_list)
@@ -439,11 +510,15 @@ class Loader():
         elif self._device_type == 3:
             # asa_m128_v2
             self._prog_time = len(self._flash_pages) * \
-                0.047 + len(self._eep_pages) * 0.05 + 0.23
+                0.047 + len(self._eep_pages) * 0.05 + 0.23 
         elif self._device_type == 4:
             # asa_m3_v1
             self._prog_time = len(self._flash_pages) * \
                 0.5 + len(self._eep_pages) * 0.05 + 0.23
+        elif self._device_type == 5:
+            #asa_m4_v1
+            self._prog_time = len(self._flash_pages) * \
+                0.5 + len(self._eep_pages) * 0.05 + 0.23 + 3
 
     def _prepare_device(self):
         """檢查裝置是否吻合設定的裝置號碼
@@ -506,6 +581,21 @@ class Loader():
                 self._flash_pages = ihex.cut_to_pages(blocks, 256)
             except Exception:
                 raise exceptions.FlashIsNotIhexError(self._flash_file)
+    
+    def _prepare_ext_flash(self):
+        """處理外部flash燒錄檔
+
+        基本操作與 _prepare_flash 無異，後續封包發送才會產生區別
+        """
+        if self._is_ext_flash_prog:
+            try:
+                blocks = ihex.parse(self._ext_flash_file)
+                self._ext_flash_size = sum([len(block['data'])
+                                        for block in blocks])
+                blocks = ihex.padding_space(blocks, 256, 0xFF)
+                self._ext_flash_pages = ihex.cut_to_pages(blocks, 256)
+            except Exception:
+                raise exceptions.FlashIsNotIhexError(self._ext_flash_file)
 
     def _prepare_eeprom(self):
         # TODO 討論 eeprom 燒錄規格，並完成
@@ -539,7 +629,11 @@ class Loader():
             time.sleep(0.03)
         elif self._protocol_version == 2:
             if self._flash_page_idx == 0:
-                self._cth.cmd_v2_flash_earse_all()
+                if self._device_type == 5:
+                    #asa_m4_v1 takes longer chip erase time
+                    self._cth.cmd_v3_flash_earse_all()
+                else :
+                    self._cth.cmd_v2_flash_earse_all()
             self._cth.cmd_v2_flash_write(address, data)
 
         self._flash_page_idx += 1
@@ -548,6 +642,35 @@ class Loader():
         if self._flash_page_idx == len(self._flash_pages):
             self._stage = next(self._stage_iter)
 
+    def _do_ext_flash_prog_step(self):
+        # 燒綠至外部 flash，實際動作內容與 flash_prog 無異
+        address = self._ext_flash_pages[self._ext_flash_page_idx]['address']
+        data = self._ext_flash_pages[self._ext_flash_page_idx]['data']
+        if self._protocol_version == 1:
+            # this command do not support protocol v1 currently 
+            pass
+        elif self._protocol_version == 2:
+            if self._ext_flash_page_idx == 0:
+                # TODO: 修改為刪除特定編號檔案
+                self._cth.cmd_v2_ext_flash_earse_all()
+            self._cth.cmd_v2_ext_flash_write(address, data)
+
+        self._ext_flash_page_idx += 1
+        self._cur_step += 1
+
+        if self._ext_flash_page_idx == len(self._ext_flash_pages):
+            self._stage = next(self._stage_iter)
+
+    def _do_ext_to_int_prog_step(self):
+        # 發送外部 flash 燒錄至內部 flash 指令
+        if self._protocol_version == 1:
+            # this command do not support protocol v1 currently
+            pass
+        elif self._protocol_version == 2:
+            # TODO: 修改為燒錄特定編號檔案
+            self._cth.cmd_v2_flash_earse_all()
+            self._cth.cmd_v2_prog_ext_to_int()
+
     def _do_eep_prog_step(self):
         if self._protocol_version == 2:
             self._cth.cmd_v2_eep_write(self._eep_pages[self._eep_page_idx])
@@ -573,5 +696,9 @@ class Loader():
             self._do_flash_prog_step()
         elif self.stage == self._Stage.EEP_PROG:
             self._do_eep_prog_step()
+        elif self.stage == self._Stage.EXT_FLASH_PROG:
+            self._do_ext_flash_prog_step()
+        elif self.stage == self._Stage.EXT_TO_INT:
+            self._do_ext_to_int_prog_step()
         elif self.stage == self._Stage.END:
             self._do_prog_end_step()
